#pragma once

#ifndef OCTREE_NODE_H
#define OCTREE_NODE_H




#include "synchronized.h"
#include "settings.h"

#include <array>

#include <stdint.h>




/*==============================================================================
	Node structure
==============================================================================*/


namespace Octree {


// **TODO: we obviously have to templatize by platform, but could we avoid templatizing by color format? parhaps by adding a base class for the texture-heap smart pointers?
template< int t_ColorFormat, int t_Platform >
struct Node {

	typename Synchronized::TextureHeap< t_ColorFormat, t_Platform >::Pointer texture;
	typename Synchronized::Heap< t_Platform >::template Pointer< Node[] > children;
};


}    // namespace Octree




/*==============================================================================
	OpenCL::Converter specialization for Node< PLATFORM_OPENCL >
==============================================================================*/


#if ( SETTING_OPENCL != 0 )


#include "opencl_converter.h"

#include <CL/cl.h>


namespace OpenCL {


template< int t_ColorFormat >
struct Converter< Octree::Node< t_ColorFormat, PLATFORM_OPENCL > > {

	struct Type {

		typename Converter< typename Synchronized::TextureHeap< t_ColorFormat, PLATFORM_OPENCL >::Pointer >::Type texture;
		typename Converter< Synchronized::Heap< PLATFORM_OPENCL >::Pointer< Octree::Node< t_ColorFormat, PLATFORM_OPENCL >[] > >::Type children;
	};

	static Type Convert( Octree::Node< t_ColorFormat, PLATFORM_OPENCL > const& source ) {

		Type result;
		result.texture = Converter< typename Synchronized::TextureHeap< t_ColorFormat, PLATFORM_OPENCL >::Pointer >::Convert( source.texture );
		result.children = Converter< Synchronized::Heap< PLATFORM_OPENCL >::Pointer< Octree::Node< t_ColorFormat, PLATFORM_OPENCL >[] > >::Convert( source.children );
		return result;
	}
};


}    // namespace OpenCL


#endif    // SETTING_OPENCL




#endif    // OCTREE_NODE_H
