require 'erb'




################################################################################
#    header_code template
################################################################################


header_code = ERB.new <<END
/*
	THIS IS A GENERATED SOURCE FILE!

	It was created from the template "<%= $0 %>" on <%= Time.new.inspect %>.
*/




#include "renderer_detail_normal_mrf_helpers.h"

#include <cmath>




namespace Renderer {


namespace detail {
END




################################################################################
#    solve_code template
################################################################################


solve_code = ERB.new <<END
/*==============================================================================
	Solve<%= name %> helper function
==============================================================================*/


void Solve<%= name %>(
	float const* const depths,
	std::pair< double, Vector< double, 2 > >* const states,
	Vector< unsigned int, 2 > const& dimension,
	Vector< unsigned int, 2 > const& coordinates,
	unsigned int const size,
	double const scale,
	double const spread,
	double const voxelSize,
	double const observationPrecision,
	double const gradientPrecision,
	double const minimumEdgeDelta,
	double const maximumEdgeDelta
)
{
	unsigned int const pitch = size * dimension[ 0 ];
	unsigned int const index = coordinates[ 1 ] * dimension[ 0 ] + coordinates[ 0 ];

	double const depth = depths[ index ];
	double const reciprocalDepth = 1 / depth;

	double state = 0;
	Vector< double, 2 > gradient = { { 0, 0 } };

	if ( reciprocalDepth > 0 ) {

		double const correctedVoxelSize = std::max( voxelSize, 2 * depth * spread );
		double const lowerEdgeDifference = minimumEdgeDelta * correctedVoxelSize;
		double const upperEdgeDifference = maximumEdgeDelta * correctedVoxelSize;

		double matrix[ 3 ][ 4 ] = {
			{ 0, 0, 0, 0 },
			{ 0, 0, 0, 0 },
			{ 0, 0, 0, 0 }
		};

		matrix[ 0 ][ 0 ] += observationPrecision;
		matrix[ 0 ][ 3 ] += observationPrecision * reciprocalDepth;
		matrix[ 1 ][ 1 ] += std::numeric_limits< double >::epsilon();
		matrix[ 2 ][ 2 ] += std::numeric_limits< double >::epsilon();
<% directions.each do |direction| %>
		{	// direction = <%= direction[ 0 ] %>,<%= direction[ 1 ] %>

			double edge = 0;    // **TODO: nonzero minimum edge weight?
			double otherState = 0;
			Vector< double, 2 > otherGradient = { { 0, 0 } };
			if ( <%=
				conditions = [];
				if ( direction[ 0 ] > 0 )
					conditions << 'coordinates[ 0 ] + size < dimension[ 0 ]'
				elsif ( direction[ 0 ] < 0 )
					conditions << 'coordinates[ 0 ] >= size'
				end
				if ( direction[ 1 ] > 0 )
					conditions << 'coordinates[ 1 ] + size < dimension[ 1 ]'
				elsif ( direction[ 1 ] < 0 )
					conditions << 'coordinates[ 1 ] >= size'
				end
				if ( conditions.length > 1 )
					conditions = conditions.map { |condition| '( ' + condition + ' )' }
				end
				conditions.join( ' && ' )
			%> ) {

				unsigned int const otherIndex = index<%= if ( direction[ 1 ] > 0 ) then ' + pitch' elsif ( direction[ 1 ] < 0 ) then ' - pitch' end %><%= if ( direction[ 0 ] > 0 ) then ' + size' elsif ( direction[ 0 ] < 0 ) then ' - size' end %>;

				double const otherDepth = depths[ otherIndex ];
				double const edgeDifference = std::abs( depth - otherDepth );
				if ( edgeDifference < upperEdgeDifference ) {

					if ( edgeDifference > lowerEdgeDifference )
						edge = ( upperEdgeDifference - edgeDifference ) / ( upperEdgeDifference - lowerEdgeDifference );
					else
						edge = 1;
				}

				otherState    = states[ otherIndex ].first;
				otherGradient = states[ otherIndex ].second;
			}

			if ( scale == 1 ) {

				double const edgeState = edge * otherState;
				double const edgeGradientPrecision = edge * gradientPrecision;

				matrix[ 0 ][ 0 ] += edge;
				<% if ( direction[ 0 ] != 0 ) %>matrix[ 0 ][ 1 ] <%= if ( direction[ 0 ] < 0 ) then '-' else '+' end %>= edge;<% else %>// matrix[ 0 ][ 1 ] unchanged<% end %>
				<% if ( direction[ 1 ] != 0 ) %>matrix[ 0 ][ 2 ] <%= if ( direction[ 1 ] < 0 ) then '-' else '+' end %>= edge;<% else %>// matrix[ 0 ][ 2 ] unchanged<% end %>
				matrix[ 0 ][ 3 ] += edgeState;

				<% if ( direction[ 0 ] != 0 ) %>matrix[ 1 ][ 0 ] <%= if ( direction[ 0 ] < 0 ) then '-' else '+' end %>= edge;<% else %>// matrix[ 1 ][ 0 ] unchanged<% end %>
				<% if ( direction[ 0 ] != 0 ) %>matrix[ 1 ][ 1 ] += edge;<% else %>// matrix[ 1 ][ 1 ] unchanged<% end %>
				matrix[ 1 ][ 1 ] += edgeGradientPrecision;
				<% if ( direction[ 0 ] != 0 ) and ( direction[ 1 ] != 0 ) then %>matrix[ 1 ][ 2 ] <%= if ( direction[ 0 ] != direction[ 1 ] ) then '-' else '+' end %>= edge;<% else %>// matrix[ 1 ][ 2 ] unchanged<% end %>
				<% if ( direction[ 0 ] != 0 ) %>matrix[ 1 ][ 3 ] <%= if ( direction[ 0 ] < 0 ) then '-' else '+' end %>= edgeState;<% else %>// matrix[ 1 ][ 3 ] unchanged<% end %>
				matrix[ 1 ][ 3 ] += edgeGradientPrecision * otherGradient[ 0 ];

				<% if ( direction[ 1 ] != 0 ) %>matrix[ 2 ][ 0 ] <%= if ( direction[ 1 ] < 0 ) then '-' else '+' end %>= edge;<% else %>// matrix[ 2 ][ 0 ] unchanged<% end %>
				<% if ( direction[ 0 ] != 0 ) and ( direction[ 1 ] != 0 ) then %>matrix[ 2 ][ 1 ] <%= if ( direction[ 0 ] != direction[ 1 ] ) then '-' else '+' end %>= edge;<% else %>// matrix[ 2 ][ 1 ] unchanged<% end %>
				<% if ( direction[ 1 ] != 0 ) %>matrix[ 2 ][ 2 ] += edge;<% else %>// matrix[ 2 ][ 2 ] unchanged<% end %>
				matrix[ 2 ][ 2 ] += edgeGradientPrecision;
				<% if ( direction[ 1 ] != 0 ) %>matrix[ 2 ][ 3 ] <%= if ( direction[ 1 ] < 0 ) then '-' else '+' end %>= edgeState;<% else %>// matrix[ 2 ][ 3 ] unchanged<% end %>
				matrix[ 2 ][ 3 ] += edgeGradientPrecision * otherGradient[ 1 ];
			}
			else {

				double const edgeState = edge * otherState;
				double const edgeGradientPrecision = edge * gradientPrecision;
				double const edgeScale = edge * scale;
				double const edgeScale2 = edgeScale * scale;
				double const edgeScaleState = edgeScale * otherState;

				matrix[ 0 ][ 0 ] += edge;
				<% if ( direction[ 0 ] != 0 ) %>matrix[ 0 ][ 1 ] <%= if ( direction[ 0 ] < 0 ) then '-' else '+' end %>= edgeScale;<% else %>// matrix[ 0 ][ 1 ] unchanged<% end %>
				<% if ( direction[ 1 ] != 0 ) %>matrix[ 0 ][ 2 ] <%= if ( direction[ 1 ] < 0 ) then '-' else '+' end %>= edgeScale;<% else %>// matrix[ 0 ][ 2 ] unchanged<% end %>
				matrix[ 0 ][ 3 ] += edgeState;

				<% if ( direction[ 0 ] != 0 ) %>matrix[ 1 ][ 0 ] <%= if ( direction[ 0 ] < 0 ) then '-' else '+' end %>= edgeScale;<% else %>// matrix[ 1 ][ 0 ] unchanged<% end %>
				<% if ( direction[ 0 ] != 0 ) %>matrix[ 1 ][ 1 ] += edgeScale2;<% else %>// matrix[ 1 ][ 1 ] unchanged<% end %>
				matrix[ 1 ][ 1 ] += edgeGradientPrecision;
				<% if ( direction[ 0 ] != 0 ) and ( direction[ 1 ] != 0 ) then %>matrix[ 1 ][ 2 ] <%= if ( direction[ 0 ] != direction[ 1 ] ) then '-' else '+' end %>= edgeScale2;<% else %>// matrix[ 1 ][ 2 ] unchanged<% end %>
				<% if ( direction[ 0 ] != 0 ) %>matrix[ 1 ][ 3 ] <%= if ( direction[ 0 ] < 0 ) then '-' else '+' end %>= edgeScaleState;<% else %>// matrix[ 1 ][ 3 ] unchanged<% end %>
				matrix[ 1 ][ 3 ] += edgeGradientPrecision * otherGradient[ 0 ];

				<% if ( direction[ 1 ] != 0 ) %>matrix[ 2 ][ 0 ] <%= if ( direction[ 1 ] < 0 ) then '-' else '+' end %>= edgeScale;<% else %>// matrix[ 2 ][ 0 ] unchanged<% end %>
				<% if ( direction[ 0 ] != 0 ) and ( direction[ 1 ] != 0 ) then %>matrix[ 2 ][ 1 ] <%= if ( direction[ 0 ] != direction[ 1 ] ) then '-' else '+' end %>= edgeScale2;<% else %>// matrix[ 2 ][ 1 ] unchanged<% end %>
				<% if ( direction[ 1 ] != 0 ) %>matrix[ 2 ][ 2 ] += edgeScale2;<% else %>// matrix[ 2 ][ 2 ] unchanged<% end %>
				matrix[ 2 ][ 2 ] += edgeGradientPrecision;
				<% if ( direction[ 1 ] != 0 ) %>matrix[ 2 ][ 3 ] <%= if ( direction[ 1 ] < 0 ) then '-' else '+' end %>= edgeScaleState;<% else %>// matrix[ 2 ][ 3 ] unchanged<% end %>
				matrix[ 2 ][ 3 ] += edgeGradientPrecision * otherGradient[ 1 ];
			}
		}
<% end %>
		// **TODO: check if matrix is singular?

		// make upper triangular with Gaussian elimination
		if ( matrix[ 1 ][ 0 ] != 0 ) {

			double const ratio = ( matrix[ 1 ][ 0 ] / matrix[ 0 ][ 0 ] );
			matrix[ 1 ][ 0 ] = 0;
			matrix[ 1 ][ 1 ] -= ratio * matrix[ 0 ][ 1 ];
			matrix[ 1 ][ 2 ] -= ratio * matrix[ 0 ][ 2 ];
			matrix[ 1 ][ 3 ] -= ratio * matrix[ 0 ][ 3 ];
		}
		if ( matrix[ 2 ][ 0 ] != 0 ) {

			double const ratio = ( matrix[ 2 ][ 0 ] / matrix[ 0 ][ 0 ] );
			matrix[ 2 ][ 0 ] = 0;
			matrix[ 2 ][ 1 ] -= ratio * matrix[ 0 ][ 1 ];
			matrix[ 2 ][ 2 ] -= ratio * matrix[ 0 ][ 2 ];
			matrix[ 2 ][ 3 ] -= ratio * matrix[ 0 ][ 3 ];
		}
		if ( matrix[ 2 ][ 1 ] != 0 ) {

			double const ratio = ( matrix[ 2 ][ 1 ] / matrix[ 1 ][ 1 ] );
			matrix[ 2 ][ 1 ] = 0;
			matrix[ 2 ][ 2 ] -= ratio * matrix[ 1 ][ 2 ];
			matrix[ 2 ][ 3 ] -= ratio * matrix[ 1 ][ 3 ];
		}

		// make diagonal with Gaussian elimination
		if ( matrix[ 1 ][ 2 ] != 0 ) {

			double const ratio = ( matrix[ 1 ][ 2 ] / matrix[ 2 ][ 2 ] );
			matrix[ 1 ][ 2 ] = 0;
			matrix[ 1 ][ 3 ] -= ratio * matrix[ 2 ][ 3 ];
		}
		if ( matrix[ 0 ][ 2 ] != 0 ) {

			double const ratio = ( matrix[ 0 ][ 2 ] / matrix[ 2 ][ 2 ] );
			matrix[ 0 ][ 2 ] = 0;
			matrix[ 0 ][ 3 ] -= ratio * matrix[ 2 ][ 3 ];
		}
		if ( matrix[ 0 ][ 1 ] != 0 ) {

			double const ratio = ( matrix[ 0 ][ 1 ] / matrix[ 1 ][ 1 ] );
			matrix[ 0 ][ 1 ] = 0;
			matrix[ 0 ][ 3 ] -= ratio * matrix[ 1 ][ 3 ];
		}

		state         = matrix[ 0 ][ 3 ] / matrix[ 0 ][ 0 ];
		gradient[ 0 ] = matrix[ 1 ][ 3 ] / matrix[ 1 ][ 1 ];
		gradient[ 1 ] = matrix[ 2 ][ 3 ] / matrix[ 2 ][ 2 ];
	}

	states[ index ].first  = state;
	states[ index ].second = gradient;
}
END




################################################################################
#    footer_code template
################################################################################


footer_code = ERB.new <<END
}    // namespace detail


}    // namespace Renderer
END




################################################################################
#    code generation
################################################################################


puts header_code.result( binding )

puts "\n\n\n\n"

name = 'Cardinal'
directions = [ [ 0, -1 ], [ 1, 0 ], [ 0, 1 ], [ -1, 0 ] ]
puts solve_code.result( binding )

puts "\n\n\n\n"

name = 'Diagonal'
directions = [ [ -1, -1 ], [ 1, -1 ], [ 1, 1 ], [ -1, 1 ] ]
puts solve_code.result( binding )

puts "\n\n\n\n"

puts footer_code.result( binding )
