/**
	\file
	\brief Contains WindowsException methods.
*/




#include "windows_exception.h"
#include "helpers.h"

#include <comdef.h>    // _com_error

#include <sstream>




/*==============================================================================
	WindowsException methods
==============================================================================*/


WindowsException::WindowsException( char const* const file, unsigned int const line, std::string const& what, HRESULT const hResult ) {

	_com_error error( hResult );

#ifdef _UNICODE
	wchar_t const* const message16 = error.ErrorMessage();
	char message[ 256 ];
	WideCharToMultiByte( CP_UTF8, 0, message16, -1, message, ARRAYLENGTH( message ), NULL, NULL );
#else    // _UNICODE
	char const* const message = error.ErrorMessage();
#endif    // _UNICODE

	std::stringstream stream;
	stream << file << " - line " << line << std::endl << '\t' << CollapseWhitespace( what ) << std::endl << '\t' << CollapseWhitespace( message );
	m_what = stream.str();
}


WindowsException::~WindowsException() throw() {
}


const char* WindowsException::what() const throw() {

	return m_what.c_str();
}
